﻿#version 400 core
out vec4 FragColor;

in vec2 TexCoords;

// Sun information
uniform vec3 viewPos;
uniform vec3 lightDir;
uniform vec3 lightColor;

// Ambient information
uniform vec3 ambientColor;
uniform float ambientIntensity;

// Screen Space Ambient Occlusion
uniform sampler2D noiseTexture;
uniform vec3 samples[32];
uniform int kernelSize;
uniform float radius;

// Shadow settings
uniform mat4 shadowMatrix;
uniform int usePixelShadows;
uniform int pixelShadowScale;
uniform float shadowBiasMax;
uniform float shadowBiasMin;

// GBuffer
uniform sampler2D gAlbedo;
uniform sampler2D gNormal;
uniform sampler2D gOcclusionRoughnessMetallic;
uniform sampler2D gPosition;

// Other textures
uniform sampler2D shadowMap;

// Position recreation
uniform sampler2D depthBuffer;

// Pixel Shadows
uniform mat4 pixelShadowTransformation;
uniform mat4 inversePixelShadowTransformation;

uniform mat4 projectionMatrix;
uniform mat4 viewMatrix;
uniform mat4 viewProjectionMatrix;
uniform mat4 projMatrixInv;
uniform mat4 viewMatrixInv;

const float PI = 3.14159265359;

float when_gt(float x, float y) {
  return max(sign(x - y), 0.0);
}

float CalculateSSAO(vec3 position, vec3 normal, vec2 texCoords, out vec3 testColor)
{
	vec3 tangent = texture(noiseTexture, texCoords * textureSize(gPosition, 0) / textureSize(noiseTexture, 0)).xyz;
	tangent = normalize(tangent - normal * dot(tangent, normal));
	mat3 tbn = mat3(tangent, cross(normal, tangent), normal);

	float occlusion = 1.0;
	float position_depth = (vec4(position, 1.0) * viewMatrix).z;
	for(int i = 0; i < kernelSize; i++)
	{
		vec4 samplePos = (vec4(position + samples[i] * radius * tbn, 1.0) * viewMatrix);

		vec4 projection_coordinates = (samplePos * projectionMatrix);

		projection_coordinates.xyz = projection_coordinates.xyz / projection_coordinates.w * 0.5 + 0.5;

		// Why doesn't this fix work..? Related to below
		// Am I missing something?
		//projection_coordinates.x = projection_coordinates.x > 1 ? 1 - (projection_coordinates.x - 1) : projection_coordinates.x;
		//projection_coordinates.x = projection_coordinates.x < 0 ? -projection_coordinates.x : projection_coordinates.x;
		//projection_coordinates.y = projection_coordinates.y > 1 ? 1 - (projection_coordinates.y - 1) : projection_coordinates.y;
		//projection_coordinates.y = projection_coordinates.y < 0 ? -projection_coordinates.y : projection_coordinates.y;

		// Fix..? Related to texture sampling
		projection_coordinates.x = min(1, max(0, projection_coordinates.x));
		projection_coordinates.y = min(1, max(0, projection_coordinates.y));

		float sample_depth = (vec4(texture(gPosition, projection_coordinates.xy).xyz, 1.0) * viewMatrix).z;

		float range_check = smoothstep(0.0, 1.0, radius / abs(position_depth - sample_depth));
		occlusion -= samplePos.z + 0.025 < sample_depth ? range_check / kernelSize : 0.0;
	}
	
	return occlusion;
}

float ShadowCalculation(vec4 fragLightSpace, vec3 normal, vec3 lightDir)
{
	// Convert frag light space to texel coordinates
	vec3 projCoords = fragLightSpace.xyz / fragLightSpace.w;
	projCoords = projCoords * 0.5 + 0.5;

	float closestDepth = texture(shadowMap, projCoords.xy).r;

	float currentDepth = projCoords.z;

	float shadow = 0;
	float bias = max(shadowBiasMax * (1.0 - dot(normal, lightDir)), shadowBiasMin);
	if(usePixelShadows == 1)
	{
		shadow = currentDepth >= closestDepth ? 1.0 : 0.0;

		if(shadow == 0.0)
		{
			vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
			for(int x = -1; x <= 1; ++x)
			{
				for(int y = -1; y <= 1; ++y)
				{
					float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r; 
					shadow += (currentDepth - bias >= pcfDepth ? 1.0 : 0.0);        
				}
			}
			shadow /= 9.0;
		}
	}
	else
	{
		vec2 texelSize = 0.5 / textureSize(shadowMap, 0);
		for(int x = -1; x <= 1; ++x)
		{
			for(int y = -1; y <= 1; ++y)
			{
				float pcfDepth = texture(shadowMap, projCoords.xy + vec2(x, y) * texelSize).r; 
				shadow += (currentDepth - bias >= pcfDepth ? 1.0 : 0.0);        
			}
		}
		shadow /= 9.0;
	}

	if(projCoords.z > 1.0)
		shadow = 0.0;

	return shadow;
}

// The below 4 functions contribute to PBR rendering
float DistributionGGX(vec3 N, vec3 H, float roughness)
{
	float a      = roughness*roughness;
	float a2     = a*a;
	float NdotH  = max(dot(N, H), 0.0);
	float NdotH2 = NdotH*NdotH;
	
	float num   = a2;
	float denom = (NdotH2 * (a2 - 1.0) + 1.0);
	denom = PI * denom * denom;
	
	return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
	float r = (roughness + 1.0);
	float k = (r*r) / 8.0;

	float num   = NdotV;
	float denom = NdotV * (1.0 - k) + k;
	
	return num / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
	float NdotV = max(dot(N, V), 0.0);
	float NdotL = max(dot(N, L), 0.0);
	float ggx2  = GeometrySchlickGGX(NdotV, roughness);
	float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
	return ggx1 * ggx2;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
	return F0 + (1.0 - F0) * pow(clamp(1.0 - cosTheta, 0.0, 1.0), 5.0);
}

void main()
{             
	// retrieve data from gbuffer
	float depth = texture(depthBuffer, TexCoords).r;
	vec3 position = texture(gPosition, TexCoords).xyz;
	vec3 Normal = texture(gNormal, TexCoords).rgb;
	vec3 Albedo = texture(gAlbedo, TexCoords).rgb;
	vec3 occlusionRoughnessMetallic = texture(gOcclusionRoughnessMetallic, TexCoords).rgb;

	vec3 testColor = vec3(0,0,0);
	float screenSpaceAmbientOcclusion = CalculateSSAO(position, Normal, TexCoords, testColor);

	// Correct gamma
	float gamma = 2.2;
	vec3 gammaCorrectedAlbedo = pow(Albedo, vec3(gamma));

	vec3 shadowPos = position;
	if(usePixelShadows == 1)
	{
		vec3 magicPos = (vec4(shadowPos, 1.0) * pixelShadowTransformation).xyz;
		double magicX = double(magicPos.x);
		double magicY = double(magicPos.y);
		double magicZ = double(magicPos.z);

		magicX = round(magicX * pixelShadowScale) / pixelShadowScale;
		magicY = round(magicY * pixelShadowScale) / pixelShadowScale;
		magicZ = round(magicZ * pixelShadowScale) / pixelShadowScale;

		magicPos = vec3(float(magicX), float(magicY), float(magicZ));//round(magicPos * pixelShadowScale) / pixelShadowScale;
		shadowPos = (vec4(magicPos, 1.0) * inversePixelShadowTransformation).xyz;
	}

	// Shadow calculation
	float shadow = ShadowCalculation(vec4(shadowPos, 1.0) * shadowMatrix, Normal, lightDir);

	// PBR
	vec3 N = normalize(Normal);
	vec3 V = normalize(viewPos - position);

	vec3 F0 = vec3(0.04);
	F0 = mix(F0, gammaCorrectedAlbedo, occlusionRoughnessMetallic.b);

	vec3 L = lightDir;
	vec3 H = normalize(V + L);

	// cook-torrance brdf
	float NDF = DistributionGGX(N, H, occlusionRoughnessMetallic.g);        
	float G = GeometrySmith(N, V, L, occlusionRoughnessMetallic.g);      
	vec3 F = fresnelSchlick(max(dot(H, V), 0.0), F0);
	
	vec3 kS = F;
	vec3 kD = vec3(1.0) - kS;
	kD *= 1.0 - occlusionRoughnessMetallic.b;	  
		
	vec3  numerator   = NDF * G * F;
	float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.0001;
	vec3  specular    = numerator / denominator;

	// add to outgoing radiance Lo
	float NdotL = max(dot(N, L), 0.0);  
	vec3 Lo = (kD * gammaCorrectedAlbedo / PI + specular) * lightColor * NdotL;

	// Output result

	// Multiply specular and diffuse by the amount of shadow and light color
	// Then add the ambient lighting based on the ambient coloring + occlusion data
	vec3 lightingOutput = (1.0 - shadow) * Lo +
		gammaCorrectedAlbedo * ambientColor * ambientIntensity * occlusionRoughnessMetallic.r * screenSpaceAmbientOcclusion;

	lightingOutput = lightingOutput / (lightingOutput + vec3(1.0));

	FragColor = vec4(pow(lightingOutput, vec3(1.0/gamma)), 1);
	gl_FragDepth = depth;

	//FragColor = vec4(testColor, 1);

	//FragColor = vec4(screenSpaceAmbientOcclusion, screenSpaceAmbientOcclusion, screenSpaceAmbientOcclusion, 1);

	// Normal Test
	//FragColor = vec4((Normal + 1) / 2, 1.0);

	// Position Test
	//FragColor = vec4(FragPos / 10, 1.0);

	// Shadow Test
	//FragColor += vec4(vec3(1 - texture(shadowMap, TexCoords).r), 1.0);

	// Depth test
	//FragColor = vec4(depth, depth, depth, 1.0);
}